<?php

namespace Nwb\ErrorHandler;

use Exception;
use Throwable;

class ErrorHandler
{
    public static $projectName = '';
    private static $show_popup_error = false;

    public function __construct(string $projectName, bool $show_popup_error = true)
    {
        if (!preg_match('/^[a-z0-9\-]+$/', $projectName)) {
            throw new Exception("Le nom du projet n'est pas valide. Veuillez consulter la doc pour obtenir des informations sur l'utilisation de ce module. ");
        }

        self::$show_popup_error = $show_popup_error;
        self::$projectName = $projectName;
    }

    /**
     * Gère les erreurs PHP et envoie un rapport de bug avec des détails sur l'erreur.
     * @param int $errno Le niveau d'erreur
     * @param string $errstr Le message d'erreur
     * @param string $errfile Le fichier dans lequel l'erreur s'est produite
     * @param int $errline La ligne à laquelle l'erreur s'est produite
     * @return array Un tableau contenant les informations sur l'erreur, y compris le code de statut, le nom, le message et la trace de débogage
     * @throws Exception Si une erreur se produit lors de l'envoi du bug à Fix
     */
    public function actionError($errno, $errstr, $errfile, $errline): array
    {
        $file = self::remove_root_from_path_ForVsCode($errfile);
        $debugBacktrace = debug_backtrace();
        $statusCode = http_response_code();
        $name = "Erreur";
        $message = '';
        $message .= "Erreur " . $errno . " : " . $errstr . "\n\n";
        $message .= 'Fichier : ' . $file . ":" . $errline . "\n";
        $message .= 'URL : ' . $this->getFullUrl() . "\n";
        $message .= "Extrait : \n" .  $this->getLinesFromFile($errfile, $errline) . "\n\n";

        $message .= self::GetCallStackForMsg($debugBacktrace);

        $message .= "\n\n" . $this->getSuperGlobals(); // Pour permettre la détection de les piratages

        // Clé de dédoublonnage pour éviter plusieurs tâches pour le même bug. Ex: $CheminDeFichierPHP + $ligneErreur + $fonctionAppelante
        $deduplicationKey = $file . $errline;

        // Génération du titre
        $title = 'Exception : ' . $file . ":" . $errline;
        $front_message = "Erreurr " . $errno . " : " . ($errstr);

        $deduplicationKey = $file . $errline;

        $url = $this->getFullUrl();

        if (self::$show_popup_error === true) {
            $this->show_error_popup($front_message, "Une erreur s'est produite");
        }

        try {
            self::sendBugToFix($message, $title, $deduplicationKey);
            $errorInfos = [
                "send_to_clickup" => true,
                "date" => date("d-m-Y H:i:s"),
                "file" => $errfile,
                "line" => $errline,
                "error_message" => $errstr,
                "error_code" => $errno,
                "url" => $url
            ];
            error_log(self::gestionJSON($errorInfos));
        } catch (Exception $e) {
            $errorInfos = [
                "send_to_clickup" => false,
                "date" => date("d-m-Y H:i:s"),
                "file" => $errfile,
                "line" => $errline,
                "error_message" => $errstr,
                "error_code" => $errno,
                "url" => $url
            ];
            error_log(self::gestionJSON($errorInfos));
        }

        return [
            'statusCode' => $statusCode,
            'name' => $name,
            'message' => $message,
            'debugBacktrace' => $debugBacktrace
        ];
    }

    /**
     * Récupère l'URL complète, y compris le protocole et le domaine.
     *
     * @return string L'URL complète
     */
    public static function getFullUrl(): string
    {
        // Récupération du domaine depuis l'url et du protocole utilisé
        $url = "";
        $domaine_site = $_SERVER['HTTP_HOST'] ?? $_SERVER['SERVER_NAME'] ?? $_SERVER['SERVER_ADDR'] ?? 'localhost';
        $protocol_site = (isset($_SERVER['REQUEST_SCHEME']) && $_SERVER['REQUEST_SCHEME'] === 'on') ? "https://" : "http://";
        $local_domains = ["::1", "fc00::/7", "127.0.0.1", "localhost"];

        if (!in_array($domaine_site, $local_domains) && strpos($domaine_site, "localhost:") !== 0) {
            $protocol_site = "https://";
        }

        $request_uri = $_SERVER['REQUEST_URI'] ?? "";
        $url = $protocol_site . $domaine_site . $request_uri;

        return $url;
    }

    /**
     * Récupère la pile d'appels pour le message d'erreur.
     *
     * Cette méthode est conçue pour traiter le résultat de `Throwable $exception->getTrace()`
     * ou de `debug_backtrace()`.
     *
     * @param array $callStack La pile d'appels à traiter.
     * @return string La pile d'appels formatée en texte.
     */
    public static function GetCallStackForMsg(array $callStack): string
    {
        $message = '';
        $message .= "Stack d'appels : \n";

        $debugBacktrace = array_map(function ($trace) {
            return array_filter($trace, function ($entry) {
                return in_array($entry, ['file', 'line']);
            }, ARRAY_FILTER_USE_KEY);
        }, $callStack);

        $message .= array_reduce($debugBacktrace, function ($carry, $item) {
            return $carry . ' -' . self::remove_root_from_path_ForVsCode(($item['file'] ?? '')) . ':' . ($item['line'] ?? '-1') . "\n";
        }, '');

        return $message;
    }

    /**
     * Gère les exceptions PHP et envoie un rapport de bug avec des détails sur l'exception.
     * @param Throwable $exception L'exception à traiter
     * @return array Un tableau contenant les informations sur l'exception, y compris le code de statut, le nom, le message et la trace de débogage
     * @throws Exception Si une erreur se produit lors de l'envoi du bug à Fix
     */
    public function actionException(Throwable $exception): array
    {
        $file = self::remove_root_from_path_ForVsCode($exception->getFile());

        $message = '';
        $message .= 'Message : ' . $exception->getMessage() . "\n";
        $message .= 'Fichier : ' . $file . ":" . $exception->getLine() . " \n";
        $message .= 'URL : ' . $this->getFullUrl() . "\n";
        $message .= "Code : " . $exception->getCode() . "\n";
        $message .= "Extrait de la ligne : \n" . $this->getLinesFromFile($exception->getFile(), $exception->getLine()) . "\n";
        $message .= "\n";

        $message .= self::GetCallStackForMsg($exception->getTrace());

        $message .= "\n\n" . $this->getSuperGlobals(); // Pour permettre la détection de les piratages

        $deduplicationKey = $file . $exception->getLine() . "Exception";
        $title = 'Exception : ' . $file . ":" . $exception->getLine();
        $front_message = "Code d'erreur : " . $exception->getCode() . " <br>
                            Message d'erreur : <br> " . $exception->getMessage() . ".";

        $url = $this->getFullUrl();

        if (self::$show_popup_error === true) {
            $this->show_error_popup($front_message, "Une erreur s'est produite");
        }

        try {
            self::sendBugToFix($message, $title, $deduplicationKey);
            $errorInfos = [
                "send_to_clickup" => true,
                "date" => date("d-m-Y H:i:s"),
                "file" => $file,
                "line" => $exception->getLine(),
                "error_message" => $exception->getMessage(),
                "error_code" => $exception->getCode(),
                "url" => $url
            ];
            error_log(self::gestionJSON($errorInfos));
        } catch (Exception $e) {
            $errorInfos = [
                "send_to_clickup" => true,
                "date" => date("d-m-Y H:i:s"),
                "file" => $file,
                "line" => $exception->getLine(),
                "error_message" => $exception->getMessage(),
                "error_code" => $exception->getCode(),
                "url" => $url
            ];
            error_log(self::gestionJSON($errorInfos));
        }

        return [
            'statusCode' => $exception->getCode(),
            'name' => 'Exception',
            'message' => $message,
            'debugBacktrace' => $exception->getTrace()
        ];
    }

    /**
     * Retourne les super globales HTTP GET, POST, FILES, REQUEST et COOKIE au format texte.
     * Cette fonction peut être utilisée pour permettre la détection de les piratages.
     * @return string Les super globales HTTP formatées en texte.
     */
    private function getSuperGlobals(): string
    {
        $output = "<SuperGlobals>\n\n";

        $superGlobals = [
            'GET' => $_GET,
            'POST' => $_POST,
            'FILES' => $_FILES,
            'REQUEST' => $_REQUEST,
            'COOKIE' => $_COOKIE,
            'SERVER' => $_SERVER
        ];

        foreach ($superGlobals as $name => $global) {
            $output .= "\$_$name = " . var_export($global, true) . ";\n\n";
        }

        $output .= "</SuperGlobals>";

        return $output;
    }

    /**
     * Gère les erreurs Yii et envoie un rapport de bug avec des détails sur l'erreur.
     * @param  Throwable|null $exception L'exception à traiter
     * @return array Un tableau contenant les informations sur l'erreur, y compris le code de statut, le nom, le message et la trace de débogage
     * @throws  \yii\base\Exception Si une erreur se produit lors de l'envoi du bug à Fix
     */
    public function actionErrorYii(?Throwable $exception)
    {
        if (!$exception) {
            return;
        }

        $statusCode = !empty($exception->getCode()) ? $exception->getCode() : 500;
        $fileFullPath = $exception->getFile();
        $file = self::remove_root_from_path_ForVsCode($fileFullPath);
        $line = $exception->getLine();
        $extraitLigne = $this->getLinesFromFile($fileFullPath, $line);
        $message = $exception->getMessage() . "\n\n";
        $message .= 'Fichier : ' . $fileFullPath . "\n";
        $message .= 'URL : ' . $this->getFullUrl() . "\n";
        $message .= 'Extrait : ' . $extraitLigne . "\n";

        // Ajout de la trace de débogage
        $errorInfo = $exception->getTrace();
        if (!empty($errorInfo)) {
            $message .= 'Stack trace : ' . "\n";

            foreach ($errorInfo as $details) {
                // Il est possible qu'il n'y ait pas de fichier, par exemple pour les fonctions anonymes
                if (empty($details['file'])) {
                    continue;
                }

                $fileFullPathStack = (isset($details['file']) ? $details['file'] : "");
                $fileStack = $fileFullPathStack;
                $fileStack = self::remove_root_from_path_ForVsCode($fileStack);
                $lineStack = (isset($details['line']) ? $details['line'] : 0);

                $message .= "Fichier : " . $fileStack . ":" . $lineStack . "\n";
                $message .= "Extrait de la ligne : \n" . $this->getLinesFromFile($fileFullPathStack, $lineStack) . "\n\n";
            }
        } else {
            $message .= "File : " . $file . ":" . $line . "\n";
            $message .= "Extrait de la ligne : \n" . $extraitLigne . "\n\n";
        }

        $message .= $this->getSuperGlobals(); // Pour permettre la détection de les piratages

        $name = "Erreur";
        $title = "Exception : " . $file . ":" . $line;
        $deduplicationKey = $file . $line;

        try {
            self::sendBugToFix($message, $title, $deduplicationKey);
        } catch (Exception $e) {
            error_log($e->getMessage());
        }

        return [
            'statusCode' => $statusCode,
            'name' => $name,
            'message' => $message,
            'debugBacktrace' => $errorInfo
        ];
    }

    /**
     * Envoie un rapport de bug à l'API de ClickUp pour être corrigé.
     * @param string $details Les détails du bug qui doivent être corrigés. Encodé en UTF-8.
     * @param string $title Le titre du bug. Encodé en UTF-8.
     * @param string $deduplicationKey La clé de dédoublonnage du bug.
     * @throws Exception Si une erreur survient lors de la création du JSON.
     */
    private function sendBugToFix(string $details, string $title, string $deduplicationKey): void
    {
        $hashDeduplicationKey = mb_strtoupper(md5($deduplicationKey)); // Creéation d'un hash pour la clé de dédoublonnage

        $data = [
            "origin" => self::$projectName,
            "title" =>  $title,
            "details" =>  $details,
            "deduplication_key" =>  $hashDeduplicationKey
        ];

        $JsonData = json_encode($data);

        if ($JsonData === false) {
            $JsonData = $this->gestionJSON($data);
            throw new Exception("Erreur lors de la création du JSON");
        }

        $ch = curl_init("https://n8n.kub.nwb.fr/webhook/to-clickup");
        curl_setopt($ch, CURLOPT_HTTPHEADER, ['Content-Type: application/json']);
        curl_setopt($ch, CURLOPT_CUSTOMREQUEST, "POST");
        curl_setopt($ch, CURLOPT_POSTFIELDS, $JsonData);
        curl_setopt($ch, CURLOPT_CRLF, true);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);

        try {
            curl_exec($ch);
        } finally {
            curl_close($ch);
        }
    }

    /**
     * Encode un tableau en JSON. Si l'encodage JSON échoue, cette fonction tente de convertir toutes les chaînes de caractères
     * du tableau en UTF-8 avant de réessayer. Si cela échoue toujours, la fonction renvoie un message d'erreur avec la date et l'heure actuelles.
     * @param array $arrayForJson Le tableau à encoder en JSON
     * @return string Le tableau encodé en JSON, ou un message d'erreur si l'encodage JSON échoue.
     */
    private function gestionJSON(array $arrayForJson): string
    {
        $jsonArray = json_encode($arrayForJson);

        if ($jsonArray === false) {
            foreach ($arrayForJson as $key => $value) {
                if (is_string($value)) {
                    $arrayForJson[$key] = mb_convert_encoding($value, 'UTF-8');
                }
            }

            $jsonArray = json_encode($arrayForJson);
        }

        if ($jsonArray === false) {
            $jsonArray = json_encode(["error_message" => "Impossible d'afficher l'erreur le " . date("d-m-Y H:i:s")]);
        }

        return $jsonArray;
    }

    /**
     * Définit les gestionnaires d'erreurs et d'exceptions à utiliser par cet objet pour php.
     * @return void
     */
    public function register(int $errorLevels = E_ALL)
    {
        set_error_handler([$this, 'actionError'], $errorLevels);
        set_exception_handler([$this, 'actionException']);
    }

    private function show_error_popup($message, $titre = "An error has occurred")
    {
        if (empty($message)) {
            return;
        }

        if (is_array($message)) {
            $message = '<ul><li>' . implode('</li><li>', $message) . '</li></ul>';
        }

        // Fonctions de gestion de la pop-up
        $popup_html = <<<HTML
<style>
#error-popup-container {
    position: fixed;
    inset: 0;
    background: rgba(0,0,0,0.5);
    display: flex;
    justify-content: center;
    align-items: center;
}
#error-popup {
    background: white;
    padding: 20px;
    border-radius: 5px;
    box-shadow: 0 0 10px rgba(0,0,0,0.5);
}
#error-close-link {
    float: right;
    color: red;
    text-decoration: none;
}
</style>
<div id="error-popup-container" onclick="this.style.display='none'">
    <div id="error-popup" onclick="event.stopPropagation()">
        <a href="#" id="error-close-link" onclick="this.closest('#error-popup-container').style.display='none'; return false;">X</a>
        <h2>{$titre}</h2>
        <p>{$message}</p>
    </div>
</div>
HTML;

        echo $this->compressHTML($popup_html);
    }

    /**
     * Compresse le HTML tout en séparant titre et message pour la lisibilité dans les endroits non interpétés en html.
     * @param string $html Le HTML à compresser.
     * @return string Le HTML compressé.
     */
    private function compressHTML(string $html): string
    {
        // Compression du HTML tout en préservant les lignes spécifiques
        $lines = explode("\n", $html);
        $compressed = '';
        foreach ($lines as $line) {
            $line = trim($line);
            if (strpos($line, '<h2>') === 0 || strpos($line, '<p>') === 0) {
                $compressed .= $line . "\n";
            } else {
                $compressed .= $line;
            }
        }
        return preg_replace('/\s+/', ' ', $compressed);
    }

    /**
     * Supprime la racine du chemin du fichier pour VS Code.
     *
     * Cette fonction prend un chemin de fichier en paramètre et supprime la partie correspondant à la racine du projet, en utilisant la valeur retournée par la méthode statique `ErrorHandler::getProjectRoot()`.
     * Elle retourne le chemin du fichier sans la partie de la racine du projet.
     * Cette fonction est principalement utilisée pour faciliter le copier-coller du chemin dans VS Code.
     *
     * @param string|null $filePath Le chemin du fichier à traiter. Peut être null.
     * @param string $rootProject La racine du projet. Si vide, la valeur retournée par la méthode statique `ErrorHandler::getProjectRoot()` est utilisée.
     * @return string Le chemin du fichier sans la partie de la racine du projet.
     */
    public static function remove_root_from_path_ForVsCode(?string $filePath, string $rootProject = ''): string
    {
        if ($filePath === null) return 'NULL';
        if (!is_string($filePath)) return 'NULL';

        if (empty($rootProject)) {
            $rootProject = self::getProjectRoot();
        }

        $root = preg_quote($rootProject, '/');
        $fileForVsCode = preg_replace('/^' . $root . '/', '', $filePath);

        // On enlève le slash du début car c'est plus pratique pour les copier coller dans VS code
        if (strlen($fileForVsCode) > 0 && $fileForVsCode[0] === '/') {
            $fileForVsCode = substr($fileForVsCode, 1);
        }

        return $fileForVsCode;
    }

    /**
     * Obtient la racine du projet en remontant jusqu'au répertoire contenant au moins un certain nombre de fichiers PHP ou Markdown.
     *
     * @param int $minFiles Le nombre minimum de fichiers (PHP ou Markdown) pour qu'un répertoire soit considéré comme la racine du projet.
     * @return string|null Chemin absolu de la racine du projet, ou null si la racine n'est pas trouvée.
     */
    public static function getProjectRoot(int $minFiles = 1): ?string
    {
        // Si une session est active et que la racine du projet est en cache, la renvoyer directement.
        if (session_status() === PHP_SESSION_ACTIVE && isset($_SESSION['project_root_cached'])) {
            return $_SESSION['project_root_cached'];
        }

        $currentPath = realpath(__DIR__);
        $projectRoot = null;

        while ($currentPath) {
            if (!self::isProjectRoot($currentPath, $minFiles)) {
                $projectRoot = $currentPath;
            } else {
                break;
            }

            $currentPath = dirname($currentPath);
        }

        // Si une session est active et que la racine du projet a été trouvée, mettre la racine du projet en cache.
        if (session_status() === PHP_SESSION_ACTIVE && $projectRoot !== null) {
            $_SESSION['project_root_cached'] = $projectRoot;
        }

        return $projectRoot;
    }

    /**
     * Vérifie si le répertoire contient au moins un certain nombre de fichiers PHP ou Markdown.
     *
     * @param string $directory Chemin absolu du répertoire à vérifier.
     * @param int $minFiles Le nombre minimum de fichiers (PHP ou Markdown) pour qu'un répertoire soit considéré comme la racine du projet.
     * @return bool Vrai si le répertoire est la racine du projet, faux sinon.
     */
    private static function isProjectRoot(string $directory, int $minFiles = 1): bool
    {
        if (strpos(strtolower($directory), '/vendor') !== false) return false;
        if (strpos(strtolower($directory), '/yii') !== false) return false;

        $fileTypes = ['*.PHP', '*.php', '*.md', '*.MD', 'composer.json', 'composer.lock', 'package.json', 'package-lock.json'];
        $totalFiles = 0;

        foreach ($fileTypes as $fileType) {
            $files = glob($directory . '/' . $fileType);
            $totalFiles += count($files);
        }

        return (!$totalFiles >= $minFiles);
    }

    /**
     * Retourne un tableau de lignes comprenant la ligne spécifiée ainsi que les lignes au-dessus et en dessous.
     *
     * @param string $filename Le nom du fichier à lire.
     * @param int $lineNumber Le numéro de la ligne souhaitée.
     * @param int $numLines (facultatif) Le nombre de lignes au-dessus et en dessous de la ligne spécifiée. Par défaut : 0.
     * @return string Le tableau de lignes sous forme de chaîne avec des retours à la ligne si la ligne spécifiée existe, sinon un message d'erreur.
     */
    public static function getLinesFromFile(string $filename, int $lineNumber, int $numLines = 0): string
    {
        $file = null;

        try {
            // fopen peut return false, mais peut aussi throw suivant la configuration / version de PHP
            $file = fopen($filename, 'r');

            if (!$file) {
                throw new \Exception("Impossible d'ouvrir le fichier : $filename");
            }
        } catch (\Exception $e) {
            return $e->getMessage();
        }

        $currentLine = 0;
        $startLine = max(1, $lineNumber - $numLines);
        $endLine = $lineNumber + $numLines;

        $lines = [];

        while (($line = fgets($file)) !== false) {
            $currentLine++;

            if ($currentLine >= $startLine && $currentLine <= $endLine) {
                $lines[] = $line;
            }

            if ($currentLine > $endLine) {
                break;
            }
        }

        fclose($file);

        if (empty($lines)) {
            return "La ligne spécifiée n'existe pas dans le fichier.";
        }

        return implode("<br/>", $lines);
    }
}
